@@ -8,6 +8,83 @@ class ImportException(Exception):
 	""" General exception for record import
 	"""
 
+class TagBreakdownModel(QtCore.QAbstractTableModel):
+	""" Model to display tag breakdown
+	"""
+	def __init__(self, parent=None):
+		super(TagBreakdownModel, self).__init__(parent=parent)
+		self.__stats = []
+
+	def rowCount(self, parent=QtCore.QModelIndex()):
+		return len(self.__stats)
+
+	def columnCount(self, parent=QtCore.QModelIndex()):
+		return 3
+
+	def updateStats(self, stats):
+		""" Stats come in as a dictionary: {tag : [money in, money out]}
+			This is stored on __stats as a list of lists [tag, in, out]
+		"""
+		for tag, (amountIn, amountOut) in stats.iteritems():
+			currentIndex = self.index(0, 0)
+			match = self.match(currentIndex, QtCore.Qt.DisplayRole, tag, 1, QtCore.Qt.MatchExactly)
+			if len(match) == 0:
+				# New item
+				row = self.rowCount()
+				self.insertRows(row, 1)
+				index = self.index(row, 0)
+				self.setData(index, tag, QtCore.Qt.EditRole)
+				index = self.index(row, 1)
+				self.setData(index, amountIn, QtCore.Qt.EditRole)
+				index = self.index(row, 2)
+				self.setData(index, -1 * amountOut, QtCore.Qt.EditRole)
+			else:
+				# Update item
+				index = self.index(match[0].row(), 1)
+				if amountIn != index.data().toPyObject():
+					self.setData(index, amountIn, QtCore.Qt.EditRole)
+
+				index = self.index(match[0].row(), 2)
+				if amountOut != index.data().toPyObject():
+					self.setData(index, -1 * amountOut, QtCore.Qt.EditRole)
+
+		# Remove old rows
+		for row in reversed(xrange(self.rowCount())):
+			index = self.index(row, 0)
+			if index.data().toString() not in stats:
+				self.removeRows(index.row(), 1)
+
+	def removeRows(self, position, rows, parent=QtCore.QModelIndex()):
+		self.beginRemoveRows(QtCore.QModelIndex(), position, position+rows-1)
+		del self.__stats[position]
+		self.endRemoveRows()
+		return True
+
+	def insertRows(self, position, rows, parent=QtCore.QModelIndex()):
+		self.beginInsertRows(QtCore.QModelIndex(), position, position + rows -1)
+		self.__stats.insert(position, [None, None, None])
+		self.endInsertRows()
+		return True
+
+	def setData(self, index, value, role=QtCore.Qt.EditRole):
+		if role == QtCore.Qt.EditRole:
+			self.__stats[index.row()][index.column()] = value
+			self.dataChanged.emit(index, index)
+			return True
+
+		return False
+
+	def data(self, index, role=QtCore.Qt.DisplayRole):
+
+		if role == QtCore.Qt.FontRole and index.column() == 0:
+			return QtGui.QFont(QtGui.QApplication.font().family(), italic=True)
+
+		elif role == QtCore.Qt.DisplayRole:
+			return self.__stats[index.row()][index.column()]
+
+		return QtCore.QVariant()
+
+
 class ImportRecord(object):
 	def __init__(self, *args):
 		super(ImportRecord, self).__init__()
@@ -52,7 +129,7 @@ class ImportModel(QtCore.QAbstractTableModel):
 			Raises ImportException on error
 		"""
 		rec = self.__records[index.row()]
-		
+
 		# Ensure we record the same timestamp for this import
 		self.__currentTimestamp = self.__currentTimestamp or QtCore.QDateTime.currentDateTime()
 
@@ -91,7 +168,7 @@ class ImportModel(QtCore.QAbstractTableModel):
 		self.dataSaved = False
 		self.dataChanged.emit(self.createIndex(0, 0), self.createIndex(self.rowCount() -1, self.columnCount() - 1))
 		self.__currentTimestamp = None
-	
+
 	def save(self):
 		""" Slot to persist changes to records 
 		"""
@@ -242,49 +319,74 @@ class RecordModel(QtSql.QSqlTableModel):
 			return flags | QtCore.Qt.ItemIsUserCheckable
 		return flags
 
-
 	def selectStatement(self):
 		if self.tableName().isEmpty():
 			return QtCore.QString()
 
 		queryFilter = self.filter()
-		queryFilter = 'AND ' + queryFilter if queryFilter else ''
+		queryFilter = 'WHERE ' + queryFilter if queryFilter else ''
 
 		query = """
-			SELECT
-				r.recordid,
-				r.checked,
-				(select count(*) from recordtags rt where rt.recordid=r.recordid),
-				r.checkdate,
-				r.date,
-				accounttypes.accountname,
-				r.description,
-				r.txdate,
-				r.amount,
-				r.insertdate,
-				r.rawdata
-			FROM records r
-			INNER JOIN accounttypes ON accounttypes.accounttypeid=r.accounttypeid
-			WHERE r.userid=%(userid)s
-			%(filter)s
-			ORDER BY r.date, r.description, r.txdate, r.recordid
+			SELECT	   r.recordid,
+					   r.checked,
+					   array_to_string(array_agg(t.tagname ORDER BY t.tagname), ','),
+					   r.checkdate,
+					   r.date,
+					   at.accountname,
+					   r.description,
+					   r.txdate,
+					   r.amount,
+					   r.insertdate,
+					   r.rawdata
+			      FROM records r
+			INNER JOIN accounttypes at ON at.accounttypeid=r.accounttypeid
+						AND r.userid=%(userid)s
+			 LEFT JOIN recordtags rt ON rt.recordid=r.recordid 
+			 LEFT JOIN tags t ON rt.tagid=t.tagid
+					   %(filter)s
+			  GROUP BY r.recordid, at.accountname
+			  ORDER BY r.date, r.description, r.txdate, r.recordid
 		""" % {'userid': self.__userId, 'filter': queryFilter}
 
 		return query
 
+	def setItemsChecked(self, indexes):
+		""" Toggles the checked status on a list of model indexes
+			The checkdate column is set to the current timestamp is
+			checked, otherwise NULL is inserted
 
-	def __getTagsforRecord(self, recordId):
+			It's more efficient to make a bulk update query than
+			set every row via the model
+		"""
+		recordIds = [self.index(index.row(), enum.kRecordColumn_RecordId).data().toPyObject() for index in indexes]
 		query = QtSql.QSqlQuery("""
-			SELECT t.tagname
-			FROM recordtags rt
-			JOIN tags t ON t.tagid=rt.tagid
-			WHERE rt.recordid=%d""" % recordId)
+			UPDATE records
+               SET checkdate = CASE WHEN checked=1 THEN NULL ELSE current_timestamp END,
+			       checked = CASE WHEN checked=1 THEN 0 ELSE 1 END
+			 WHERE recordid in (%s)
+			""" % ','.join(str(rec) for rec in recordIds))
 
-		tagNames = []
-		while query.next():
-			tagNames.append(str(query.value(0).toString()))
+		if query.lastError().isValid():
+			return False
 
-		return ', '.join(tagNames)
+		self.dataChanged.emit(indexes[0], indexes[-1])
+
+		return True
+
+	def deleteRecords(self, indexes):
+		recordIds = [self.index(index.row(), enum.kRecordColumn_RecordId).data().toPyObject() for index in indexes]
+		query = QtSql.QSqlQuery("""
+			DELETE FROM records 
+			      WHERE recordid in (%s)
+		""" % ','.join(str(rec) for rec in recordIds))
+
+		if query.lastError().isValid():
+			return False
+
+		self.select()
+		self.dataChanged.emit(indexes[0], indexes[-1])
+
+		return True
 
 	def data(self, item, role=QtCore.Qt.DisplayRole):
 		""" Return data from the model, formatted for viewing
@@ -301,7 +403,7 @@ class RecordModel(QtSql.QSqlTableModel):
 		if role == QtCore.Qt.ToolTipRole:
 			if item.column() == enum.kRecordColumn_Tags:
 				#Show tag names for this record
-				return self.__getTagsforRecord(self.data(self.index(item.row(), enum.kRecordColumn_RecordId)).toPyObject())
+				return self.record(item.row()).value(enum.kRecordColumn_Tags).toString()
 
 			elif item.column() == enum.kRecordColumn_Checked:
 				if super(RecordModel, self).data(self.index(item.row(), enum.kRecordColumn_Checked)).toBool():
@@ -328,10 +430,9 @@ class RecordModel(QtSql.QSqlTableModel):
 		if role == QtCore.Qt.DecorationRole:
 			if item.column() == enum.kRecordColumn_Tags:
 				# Show tag icon if we have any
-				if super(RecordModel, self).data(item).toPyObject() > 0:
+				if super(RecordModel, self).data(item).toPyObject():
 					return QtGui.QIcon(':/icons/tag_yellow.png')
 
-
 		if role == QtCore.Qt.DisplayRole:
 			if item.column() in (enum.kRecordColumn_Checked, enum.kRecordColumn_Tags):
 				# Don't display anything for these fields
@@ -344,8 +445,8 @@ class RecordModel(QtSql.QSqlTableModel):
 					return QtCore.QString.number(abs(val), 'f', 2)
 
 			elif item.column() == enum.kRecordColumn_Description:
-				# Truncate the description field to 30 chars, and replace multiple spaces with single
-				return super(RecordModel, self).data(item).toString().replace(QtCore.QRegExp('[ ]+'), ' ').left(30)
+				# Replace multiple spaces with single
+				return super(RecordModel, self).data(item).toString().replace(QtCore.QRegExp('[ ]+'), ' ')
 
 			elif item.column() == enum.kRecordColumn_Date:
 				return super(RecordModel, self).data(item).toDate()
@@ -362,19 +463,7 @@ class RecordModel(QtSql.QSqlTableModel):
 		""" Save new checkstate role changes in database
 		"""
 		if role == QtCore.Qt.CheckStateRole and index.column() == enum.kRecordColumn_Checked:
-			if value.toPyObject() == QtCore.Qt.Checked:
-				dbvalue = QtCore.QVariant(1)
-				timestamp = QtCore.QVariant(QtCore.QDateTime.currentDateTime())
-			else:
-				dbvalue = QtCore.QVariant(0)
-				timestamp = QtCore.QVariant()
-
-			checkDateIndex = self.index(index.row(), enum.kRecordColumn_CheckDate)
-
-			# Save the new state and timestamp
-			if super(RecordModel, self).setData(index, dbvalue) and \
-				super(RecordModel, self).setData(checkDateIndex, timestamp):
-				return self.submitAll()
+			return self.setItemsChecked([index])
 
 		return False
 
@@ -412,6 +501,7 @@ class TagModel(QtSql.QSqlTableModel):
 		model = QtSql.QSqlTableModel(self)
 		model.setTable('recordtags')
 		model.select()
+		self.__tagCache = {}
 
 		self.__recordTagModel = model
 
@@ -451,14 +541,19 @@ class TagModel(QtSql.QSqlTableModel):
 		return super(TagModel, self).data(item, role)
 
 	def __getRecordIdsForTag(self, tagId):
-		""" Generator to get all recordIds assigned to a given tag (id)
+		""" Returns all recordIds assigned to a given tag (id)
 		"""
-		for row in xrange(self.__recordTagModel.rowCount()):
-			recordId = self.__recordTagModel.index(row, enum.kRecordTagsColumn_RecordId).data().toPyObject()
-			recordTagId = self.__recordTagModel.index(row, enum.kRecordTagsColumn_TagId).data().toPyObject()
+		if tagId not in self.__tagCache:
+			recordIds = []
+			for row in xrange(self.__recordTagModel.rowCount()):
+				recordId = self.__recordTagModel.index(row, enum.kRecordTagsColumn_RecordId).data().toPyObject()
+				recordTagId = self.__recordTagModel.index(row, enum.kRecordTagsColumn_TagId).data().toPyObject()
 
-			if tagId == recordTagId:
-				yield recordId
+				if tagId == recordTagId:
+					recordIds.append(recordId)
+			self.__tagCache[tagId] = recordIds
+
+		return self.__tagCache[tagId]
 
 	def setData(self, index, value, role=QtCore.Qt.EditRole):
 		""" Handle checkstate role changes 
@@ -492,6 +587,11 @@ class TagModel(QtSql.QSqlTableModel):
 			self.__recordTagModel.removeRows(row, 1)
 			self.__recordTagModel.submit()
 
+		self.__removeTagFromCache(tagId)
+
+	def __removeTagFromCache(self, tagId):
+		if tagId in self.__tagCache:
+			self.__tagCache.pop(tagId)
 
 	def __addTagsToRecords(self, tagId):
 		""" Assign the tag to our records
@@ -500,16 +600,15 @@ class TagModel(QtSql.QSqlTableModel):
 		newRecordIds = self.__recordIds - currentIds
 
 		row = self.__recordTagModel.rowCount()
-		self.__recordTagModel.insertRows(row, 1)
 
 		for recordId in newRecordIds:
 			# Only one row at a time can be inserted when using the OnFieldChange or OnRowChange update strategies.
 			self.__recordTagModel.insertRows(row, 1)
 			self.__recordTagModel.setData(self.__recordTagModel.index(row, enum.kRecordTagsColumn_RecordId), QtCore.QVariant(recordId))
 			self.__recordTagModel.setData(self.__recordTagModel.index(row, enum.kRecordTagsColumn_TagId), QtCore.QVariant(tagId))
-
 			self.__recordTagModel.submit()
 
+		self.__removeTagFromCache(tagId)
 
 	def addTag(self, tagName):
 		""" Add a new tag and assign to our current records
@@ -597,7 +696,7 @@ class CheckComboModel(QtSql.QSqlTableModel):
 		self.__userRoleColumn = None
 
 	def setUserRoleColumn(self, column):
-		""" Set the database table colum to use for UserRole data
+		""" Set the database table column to use for UserRole data
 		"""
 		self.__userRoleColumn = column
 
